@@ -14,20 +14,24 @@ Stéphane Bortzmeyer <stephane+frama@bortzmeyer.org>
 """
 
 # WARNING: if you modify it here, also change setup.py https://packaging.python.org/guides/single-sourcing-package-version/#single-sourcing-the-version
-VERSION = '1.1.10'
+VERSION = '2.1.1'
 
 import os
 import json
 import time
 import urllib.request, urllib.error, urllib.parse
 import random
+import re
 import copy
 import sys
 import getopt
 import string
+import enum
+import socket
 
 authfile = "%s/.atlas/auth" % os.environ['HOME']
 base_url = "https://atlas.ripe.net/api/v2/measurements"
+base_url_probes = "https://atlas.ripe.net/api/v2/probes"
 
 # The following parameters are currently not settable. Anyway, be
 # careful when changing these, you may get inconsistent results if you
@@ -53,7 +57,10 @@ class AuthFileEmpty(Exception):
     pass
 
 class RequestSubmissionError(Exception):
-    pass
+    def __init__(self, http_status, reason, body):
+        self.status = http_status
+        self.reason = reason
+        self.body = body
 
 class FieldsQueryError(Exception):
     pass
@@ -64,6 +71,12 @@ class MeasurementNotFound(Exception):
 class MeasurementAccessError(Exception):
     pass
 
+class ProbeNotFound(Exception):
+    pass
+
+class ProbeAccessError(Exception):
+    pass
+
 class ResultError(Exception):
     pass
 
@@ -77,6 +90,36 @@ class InternalError(Exception):
 class WrongAssumption(Exception):
     pass
 
+# Utilities
+def format_error(error):
+    msg = ""
+    if error.status is not None:
+        msg += "HTTP status %s " % error.status
+    msg += error.reason
+    if error.body is not None and error.body != "":
+        msg += " %s" % error.body
+    return msg
+
+Host_Type = enum.Enum('Host-Type', ['IPv6', 'IPv4', 'Name'])
+def host_type(s):
+    """Takes an argument which is the identifier for a host and returns a host type. """
+    try:
+        addr = socket.inet_pton(socket.AF_INET6, s)
+        return Host_Type.IPv6
+    except socket.error: # not a valid IPv6 address
+        try:
+            addr = socket.inet_pton(socket.AF_INET, s) # Note that it
+            # fails with unusual but common syntaxes such as raw
+            # integers.
+            return Host_Type.IPv4
+        except socket.error: # not a valid IPv4 address either
+            return Host_Type.Name
+
+def clean_key (url):
+    """Removes key from URL, for safer error messages."""
+    (url, key) = re.split("\?key=", url)
+    return url
+
 class Config:
     def __init__(self):
         # Default values
@@ -94,8 +137,12 @@ class Config:
         self.machine_readable = False
         self.measurement_id = None
         self.display_probes = False
+        self.display_probe_asns = False
+        self.cache_probes = None
         self.ipv4 = False
         self.private = False
+        self.resolve_on_probe = False
+        self.tags = None # Tags we send to tag the measurement, not tags of the probes
         self.port = 80
         self.size = 64
         self.spread = None
@@ -110,6 +157,8 @@ class Config:
         --verbose or -v : makes the program more talkative
         --help or -h : this message
         --displayprobes or -o : display the probes numbers (WARNING: big lists)
+        --displayprobeasns : display the (unique) probe ASNumbers (currently only for blaeu-resolve)
+        --cache-probes=<file.json> : cache probe data
         --country=2LETTERSCODE or -c 2LETTERSCODE : limits the measurements to one country (default is world-wide)
         --area=AREACODE or -a AREACODE : limits the measurements to one area such as North-Central (default is world-wide)
         --asn=ASnumber or -n ASnumber : limits the measurements to one AS (default is all ASes)
@@ -118,27 +167,34 @@ class Config:
         --requested=N or -r N : requests N probes (default is %s)
         --percentage=X or -p X : stops the program as soon as X %% of the probes reported a result (default is %s %%)
         --measurement-ID=N or -m N : do not start a measurement, just analyze a former one
-        --old_measurement MSMID or -g MSMID : uses the probes of measurement MSMID
+        --old-measurement MSMID or -g MSMID : uses the probes of measurement MSMID
         --include TAGS or -i TAGS : limits the measurements to probes with these tags (a comma-separated list)
         --exclude TAGS or -e TAGS : excludes from measurements the probes with these tags (a comma-separated list)
         --port=N or -t N : destination port for TCP (default is %s)
         --size=N or -z N : number of bytes in the packet (default is %s bytes)
         --ipv4 or -4 : uses IPv4 (default is IPv6, except if the parameter or option is an IP address, then it is automatically found)
+        --tags TAGS : tag the measurement (no relationship with probes tags) (a comma-separated list)
         --spread or -w : spreads the tests (add a delay before the tests)
         --private : makes the measurement private
-        --machinereadable or -b : machine-readable output, to be consumed by tools like grep or cut
+        --resolve-on-probe : resolve names with the probe's DNS resolver
+        --machine-readable or -b : machine-readable output, to be consumed by tools like grep or cut
         """ % (self.requested, int(self.percentage_required*100), self.port, self.size), file=sys.stderr)
 
     def parse(self, shortOptsSpecific="", longOptsSpecific=[], parseSpecific=None, usage=None):
         if usage is None:
             usage = self.usage
         try:
+            # We keep some old syntaxes that were legal in the past
+            # such as --old_measurement (underscore) or
+            # --machineredable).
             optlist, args = getopt.getopt (sys.argv[1:],
-                                           "4a:bc:e:f:g:hi:m:n:op:r:s:t:vw:z:" + shortOptsSpecific,
-                                           ["requested=", "country=", "area=", "asn=", "prefix=", "probes=",
+                                           "4a:bc:e:f:g:hi:m:n:op:r:s:t:uvw:z:" + shortOptsSpecific,
+                                           ["requested=", "country=", "area=", "asn=", "prefix=", "cache-probes=", "probes=",
                                             "port=", "percentage=", "include=", "exclude=", "version",
-                                            "measurement-ID=", "old_measurement=", "displayprobes", "size=",
-                                            "ipv4", "private", "machinereadable", "spread=", "verbose", "help"] +
+                                            "measurement-ID=", "old-measurement=", "old_measurement=",
+                                            "display-probes", "displayprobes", "displayprobeasns", "size=",
+                                            "ipv4", "private", "resolve-on-probe",
+                                            "machine-readable", "machinereadable", "spread=", "verbose", "tags=", "help"] +
                                            longOptsSpecific)
             for option, value in optlist:
                 if option == "--country" or option == "-c":
@@ -149,6 +205,8 @@ class Config:
                     self.asn = value
                 elif option == "--prefix" or option == "-f":
                     self.prefix = value
+                elif option == "--cache-probes":
+                    self.cache_probes = value
                 elif option == "--probes" or option == "-s":
                     self.probes = value # Splitting (and syntax checking...) delegated to Atlas
                 elif option == "--percentage" or option == "-p":
@@ -160,7 +218,7 @@ class Config:
                     self.port = int(value)
                 elif option == "--measurement-ID" or option == "-m":
                     self.measurement_id = value
-                elif option == "--old_measurement" or option == "-g":
+                elif option == "--old-measurement" or option == "--old_measurement" or option == "-g":
                     self.old_measurement = value
                 elif option == "--verbose" or option == "-v":
                     self.verbose = True
@@ -168,18 +226,24 @@ class Config:
                     self.ipv4 = True
                 elif option == "--private":
                     self.private = True
+                elif option == "--resolve-on-probe" or option == "-u":
+                    self.resolve_on_probe = True
                 elif option == "--size" or option == "-z":
                     self.size = int(value)
                 elif option == "--spread" or option == "-w":
                     self.spread = int(value)
-                elif option == "--displayprobes" or option == "-o":
+                elif option == "--display-probes" or  option == "--displayprobes" or option == "-o":
                     self.display_probes = True
+                elif option == "--displayprobeasns":
+                    self.display_probe_asns = True
                 elif option == "--exclude" or option == "-e":
                     self.exclude = value.split(",")
                 elif option == "--include" or option == "-i":
-                    # TODO allows to specify stable probes https://labs.ripe.net/Members/chris_amin/new-ripe-atlas-probe-stability-system-tags
+                    # See the file TAGS
                     self.include = value.split(",")
-                elif option == "--machinereadable" or option == "-b":
+                elif option == "--tags":
+                    self.tags = value.split(",")
+                elif option == "--machine-readable" or option == "--machinereadable" or option == "-b":
                     self.machine_readable = True
                 elif option == "--help" or option == "-h":
                     usage()
@@ -224,17 +288,16 @@ class Config:
             if not self.default_requested:
                 print("Warning: --requested=%d ignored since a list of probes was requested" % self.requested, file=sys.stderr)
         if self.old_measurement is not None:
-            if self.country is not None:
-                print("Warning: --country ignored since we use probes from a previous measurement", file=sys.stderr)
-            if self.area is not None:
-                print("Warning: --area ignored since we use probes from a previous measurement", file=sys.stderr)
-            if self.prefix is not None:
-                print("Warning: --prefix ignored since we use probes from a previous measurement", file=sys.stderr)
-            if self.asn is not None:
-                print("Warning: --asn ignored since we use probes from a previous measurement", file=sys.stderr)
-            if self.probes is not None:
-                print("Warning: --probes ignored since we use probes from a previous measurement", file=sys.stderr)
-            # TODO include and exclude should trigger a similar warning...
+            def ignored(variable, name):
+                if variable is not None:
+                    print("Warning: --%s ignored since we use probes from a previous measurement" % name, file=sys.stderr)
+            ignored(self.country, "country")
+            ignored(self.area, "area")
+            ignored(self.prefix, "prefix")
+            ignored(self.asn, "asn")
+            ignored(self.probes, "probes")
+            ignored(self.include, "include")
+            ignored(self.exclude, "exclude")
         if self.probes is not None:
             self.requested = len(self.probes.split(","))
         data = { "is_oneoff": True,
@@ -243,10 +306,11 @@ class Config:
                  "probes": [
                      {"requested": self.requested} ] }
         if self.old_measurement is not None:
-            data["probes"][0]["requested"] = 500 # Dummy value, anyway,
-                                                    # but necessary to get
-                                                    # all the probes
-            # TODO: the huge value of "requested" makes us wait a very long time
+            old_measurement = Measurement(data=None, id=self.old_measurement)
+            data["probes"][0]["requested"] = old_measurement.num_probes 
+            if self.verbose:
+                print("Using %i probes from measurement #%s" % \
+                      (data["probes"][0]["requested"], self.old_measurement))
             data["probes"][0]["type"] = "msm"
             data["probes"][0]["value"] = self.old_measurement
             data["definitions"][0]["description"] += (" from probes of measurement #%s" % self.old_measurement)
@@ -280,8 +344,12 @@ class Config:
             data["definitions"][0]['af'] = 6 
         if self.private:
             data["definitions"][0]['is_public'] = False
+        if self.resolve_on_probe:
+            data["definitions"][0]['resolve_on_probe'] = True
         if self.size is not None:
-            data["definitions"][0]['size'] = self.size    
+            data["definitions"][0]['size'] = self.size
+        if self.tags is not None:
+            data["definitions"][0]["tags"] = self.tags
         if self.spread is not None:
             data["definitions"][0]['spread'] = self.spread
         data["probes"][0]["tags"] = {}
@@ -325,7 +393,7 @@ class Measurement():
         """
 
         if data is None and id is None:
-            raise RequestSubmissionError("No data and no measurement ID")
+            raise RequestSubmissionError(None, "No data and no measurement ID", None)
         
         # TODO: when creating a dummy measurement, a key may not be necessary if the measurement is public
         if not key:
@@ -364,12 +432,9 @@ class Measurement():
                 self.id = results["measurements"][0]
                 conn.close()
             except urllib.error.HTTPError as e:
-                raise RequestSubmissionError("Status %s, reason \"%s : %s\"" % \
-                                             (e.code, e.reason, e.read()))
+                raise RequestSubmissionError(e.code, e.reason, e.read().decode())
             except urllib.error.URLError as e:
-                raise RequestSubmissionError("Reason \"%s\"" % \
-                                             (e.reason))
-
+                raise RequestSubmissionError(None, e.reason, "URL %s" % clean_key(self.url))
 
             self.gen = random.Random()
             self.time = time.gmtime()
@@ -527,3 +592,95 @@ class Measurement():
             except urllib.error.URLError as e:
                 raise ResultError(e.reason)
         return result_data
+
+class Probe():
+    """ An Atlas probe, identified by its ID (such as #1010569) in the field "id" """
+    __probe_attributes = {
+        "id": "id",
+        "status": "status.name",
+        "address_v4": "address_v4",
+        "address_v6": "address_v6",
+        "asn_v4": "asn_v4",
+        "asn_v6": "asn_v6",
+        "country_code": "country_code",
+        "is_public": "is_public",
+        "prefix_v4": "prefix_v4",
+        "prefix_v6": "prefix_v6",
+        "tags": "tags",
+        "description": "description",
+        "geometry": "geometry",
+        "first_connected": "first_connected",
+        "last_connected": "last_connected",
+    }
+
+    def __init__(self, id, fetch=True):
+        if not fetch:
+            self.id = id
+            return
+        self.url = base_url_probes + "/%s" % id
+        try:
+            conn = urllib.request.urlopen(JsonRequest(self.url))
+        except urllib.error.HTTPError as e:
+            if e.code == 404:
+                raise ProbeNotFound
+            else:
+                raise ProbeAccessError("HTTP %s, %s %s" % (e.code, e.reason, e.read()))
+        except urllib.error.URLError as e:
+            raise ProbeAccessError("Reason \"%s\"" % \
+                                         (e.reason))
+        probe_data = json.loads(conn.read().decode('utf-8'))
+        def nested_get(d, keys):
+            k = keys.pop(0)
+            d = d[k]
+            return nested_get(d, keys) if keys else d
+        for key, lookup_key in self.__probe_attributes.items():
+            setattr(self, key, nested_get(probe_data, lookup_key.split(".")))
+
+    def __str__(self):
+        return "Probe #%s, %s" % (self.id, self.description)
+
+    def __repr__(self):
+        return "Probe #%s, %s" % (self.id, self.description)
+
+    def __eq__(self, other):
+        return self.__dict__ == other.__dict__
+
+    @classmethod
+    def from_dict(cls, data):
+        """Creates a Probe object from a dictionary (not a JSON string)"""
+        if not all(k in data for k in cls.__probe_attributes):
+            return
+        probe = cls(data["id"], fetch=False)
+        for key, value in data.items():
+            setattr(probe, key, value)
+        return probe
+
+class ProbeCache():
+    """ A cache for probe data, to avoid querying the Atlas servers each time. """
+    __cache__ = {}
+
+    def __init__(self, filename):
+        self.filename = filename
+        self.data = {} if not os.path.exists(filename) else \
+            json.load(open(filename), object_hook=lambda o: Probe.from_dict(o) or o)
+
+    def __str__(self):
+        return "ProbeCache %s, %s probes" % (self.filename, len(self.data))
+
+    def __repr__(self):
+        return "ProbeCache %s, %s probes" % (self.filename, len(self.data))
+
+    def get(self, probe):
+        return self.data.get(str(probe.id))
+
+    def put(self, probe):
+        prev_data = self.data.copy()
+        self.data[str(probe.id)] = probe
+        if self.data != prev_data:
+            json.dump(self.data, open(self.filename, "w"), default=lambda o: o.__dict__)
+        return probe
+
+    @classmethod
+    def cache_probe_id(cls, filename, id):
+        cache = cls.__cache__.setdefault(filename, cls(filename))
+        return cache.get(Probe(id)) or cache.put(Probe(id))
